@@ -15,15 +15,31 @@
  *                                                                         *
  ***************************************************************************/
 
+/* 1 bedeutet, dass manche Signale abgefangen werden, um die Log-Datei
+   abzuspeichern. */
+#define signale_fangen 1
+
+
+
 #include <cstdlib>
+#include <cstdio>
+
+#if signale_fangen
+#include <csignal>
+#endif
 
 #include <qapplication.h>
 #include <qaccel.h>
 
+
 #include "cuyointl.h"
-#include "prefs.h"
-#include "startatdlg.h"
+#include "aufnahme.h"
+#include "prefsdaten.h"
 #include "kiplayer.h"
+#include "spielfeld.h"
+#include "fehler.h"
+#include "punktefeld.h"
+#include "global.h"
 
 #include "cuyo.h"
 
@@ -40,22 +56,85 @@
 #define MENU_KI_player 133
 
 /* # Zeitschritte, die nach der Zeitbonus-Animation noch gewartet wird */
-#define nachbonus_wartezeit 20
+#define nachbonus_wartezeit 50
 
-/* Nur zur bergabe an setSpielerZahl */
-#define spz_ki 3
 
-/* Wird in main.cpp definiert */
-extern bool gDebug;
+/*** Defines fr Fenster-Layout ***/
+/* Hhe von dem Bereich ber den Spielfeldern mit der Punkteanzeige */
+#define L_oben_hoehe 32
+/* Rand zwischen den verschiedenen Objekten */
+#define L_rand 4
+/* Fensterbreite */
+#define L_fenster_breite (L_rand + (grx * gric + L_rand) * max_spielerzahl)
+
+
+
+#if signale_fangen
+
+void signalHandler(int s) {
+  try {
+    Aufnahme::speichern(getLogName());
+    /* send_log_string ist in fehler.h definiert. */
+    fprintf(stderr, "%s\n", send_log_string);
+  } catch (Fehler) {}
+  
+  /* Braucht man das? Geht das? */
+  raise(s);
+}
+
+
+void signaleAn() {
+  /*{
+    struct sigaction act; 
+    act.sa_handler = speicherFehler;
+    act.sa_mask = 0;
+    act.sa_flags = SA_ONESHOT;
+    sigaction(SIGSEGV, &act, 0);
+  }*/
+  signal(SIGILL, signalHandler);  // illegal instruction
+  signal(SIGSEGV, signalHandler); // segmentation fault
+  signal(SIGFPE, signalHandler);  // floating point exception
+}
+
+void signaleAus() {
+  signal(SIGILL, SIG_DFL);  // illegal instruction
+  signal(SIGSEGV, SIG_DFL); // segmentation fault
+  signal(SIGFPE, SIG_DFL);  // floating point exception
+}
+
+#else
+void signaleAn() {
+}
+
+void signaleAus() {
+}
+#endif
+
+
+
+Cuyo * Cuyo::gCuyo;
+
 
 
 Cuyo::Cuyo(QWidget*parent,const char* name):
   QWidget(parent,name),
-  mZeitlupe(false), mZaehlerZeitlupe(0), mRueberReihenTest(0), mFlag(0),
-  mPauseBild(0)
+  mEinzelschritt(false),
+  mZeitlupe(false), mZaehlerZeitlupe(0), mRueberReihenTest(0),
+  mAbspielen(0), mSchnellvorlauf(0),
+  mPauseBild(0), mTimer(0)
 {
+  /** Uns selbst in die globale Variable speichern. */
+  gCuyo = this;
+  
+  /* Preferences-Objekt erzeugen. Hier wird die .cuyo-Datei gelesen.
+     PrefsDaten speichert sich selbst in die globale Variable gPrefs. */
+  new PrefsDaten();
 
-  ld = new LevelDaten();
+  /* Hier findet das parsen statt. (Fehler werden im try-catch-Block
+     der main()-Routine ausgegeben.) Das LevelDaten-Objekt weist
+     sich selber der globalen Variable ld zu. */
+  new LevelDaten();
+  
   
   /* Fenstergre schon mal provisorisch einstellen, damit die Menleiste
      die richtige Breite kriegt */
@@ -68,26 +147,26 @@ Cuyo::Cuyo(QWidget*parent,const char* name):
   /* Fenstergre richtig einstellen */
   int L_oben = L_rand + mMenuBar->height();
   int L_fenster_hoehe = gry*gric + L_oben + 2 * L_rand + L_oben_hoehe;
-  int i;
   resize(L_fenster_breite, L_fenster_hoehe);
   setMinimumSize(L_fenster_breite, L_fenster_hoehe);
   setMaximumSize(L_fenster_breite, L_fenster_hoehe);
 	
   /* Fenster mit diversem Zeug fllen... */
-		
-  /* Punktelabel */
-  for (i = 0; i < max_spielerzahl; i++) {
-    punkteLabel[i] = new QLabel(this);
-    punkteLabel[i]->resize(L_graslabel_breite, L_oben_hoehe);
+
+  /* Punktelabel. Hoffentlich kommen die nicht auf die Idee, sich
+     jetzt gleich zu malen. Das geht nmlich erst, wenn wir
+     ld->ladKeinenLevel() aufgerufen haben. Und das wiederum geht
+     erst nach dem Laden des Titelbilds. */
+  for (int i = 0; i < max_spielerzahl; i++) {
+    mPunkteFeld[i] = new Punktefeld(this);
+    mPunkteFeld[i]->resize(grx * gric, L_oben_hoehe);
   }
-	
-  punkteLabel[0]->move(L_rand, L_oben);
-  punkteLabel[0]->setAlignment(AlignVCenter | AlignLeft);
-  punkteLabel[1]->move(L_fenster_breite - L_rand - L_graslabel_breite, L_oben);
-  punkteLabel[1]->setAlignment(AlignVCenter | AlignRight);
+
+  mPunkteFeld[0]->move(L_rand, L_oben);
+  mPunkteFeld[1]->move(L_fenster_breite - L_rand - grx * gric, L_oben);
 	
 
-  for (i = 0; i < max_spielerzahl; i++) {
+  for (int i = 0; i < max_spielerzahl; i++) {
     /* Spielfeld erzeugen */
     mSpielfeld[i] = new Spielfeld(i > 0, this); // (i > 0) = rechter Spieler
     mSpielfeld[i]->move(L_rand + (grx*gric+L_rand)*i,
@@ -99,17 +178,20 @@ Cuyo::Cuyo(QWidget*parent,const char* name):
       mSpielfeld[i]->erzeugTitelbild();
     } catch (Fehler f) {
       fprintf(stderr, _("Could not load Title:\n"));
-      fprintf(stderr, "%s\n", f.mText.data());
+      fprintf(stderr, "%s\n", f.getText().data());
     }
   } // for i
 
+  /* Punkte-Icons laden und Schriftfarbe setzen. */
+  ld->ladKeinenLevel();
+
   /* Pause-Bild laden */
   try {
     mPauseBild = new Bilddatei();
     mPauseBild->laden("pause.xpm");
   } catch (Fehler f) {
     fprintf(stderr, _("Could not load pause.xpm:\n"));
-    fprintf(stderr, "%s\n", f.mText.data());
+    fprintf(stderr, "%s\n", f.getText().data());
     mPauseBild = 0;
   }
 
@@ -117,7 +199,7 @@ Cuyo::Cuyo(QWidget*parent,const char* name):
   for (int i = 0; i < max_spielerzahl; i++) {
     __String s;
     /** VERSION wird in config.h definiert */
-    s.sprintf(_("Version %s\n%s\nPlayer %d\n\n"),
+    s.sprintf(_("By Immi\nVersion %s\n%s\nPlayer %d\n\n"),
 	      VERSION,
 	      gDebug ? _("Debug-mode (h = help)\n") : "",
 	      i + 1);
@@ -128,14 +210,12 @@ Cuyo::Cuyo(QWidget*parent,const char* name):
  	
     connect(mSpielfeld[i], SIGNAL(sigBekommPunkte(bool, int)),
 	    SLOT(neuePunkte(bool, int)));
-    connect(mSpielfeld[i], SIGNAL(sigGrasGeplatzt()),
-	    SLOT(grasGeplatzt()));
     for (int j = 0; j < max_spielerzahl; j++)
       if (i != j) {
 	connect(mSpielfeld[i], SIGNAL(sigSendeGraue(int)),
 		mSpielfeld[j], SLOT(empfangeGraue(int)));
-	connect(mSpielfeld[i], SIGNAL(sigWillReihe(int, bool &)),
-		mSpielfeld[j], SLOT(gebReihe(int, bool &)));
+	connect(mSpielfeld[i], SIGNAL(sigWillReihe(int, int &)),
+		mSpielfeld[j], SLOT(gebReihe(int, int &)));
 	connect(mSpielfeld[i], SIGNAL(sigWillStein(Blop &)),
 		mSpielfeld[j], SLOT(gebStein(Blop &)));
       }
@@ -145,8 +225,6 @@ Cuyo::Cuyo(QWidget*parent,const char* name):
   /* KI-Spieler erzeugen */
   mKI = new KIPlayer(mSpielfeld[1]);
 
-  liesPreferences();	
-	
   setSpielerZahl(1); // Sollte erst nach Erzeugen der Mens aufgerufen werden...
 
   /* Damit Weiterspielen grad nicht geht. */
@@ -160,6 +238,7 @@ Cuyo::Cuyo(QWidget*parent,const char* name):
 Cuyo::~Cuyo(){
   delete ld;
   delete mKI;
+  delete gPrefs;
 }
 
 /**  */
@@ -223,59 +302,18 @@ void Cuyo::MenuAufruf(int id) {
     break;
 			
   case MENU_restart_level:
-    ASSERT(mLevelNr != 0);
+    CASSERT(mLevelNr != 0);
     startSpiel();
     break;
 			
   case MENU_start_at: {
-    StartAtDlg sa(this);
-						
-    /* Liste der Level basteln, die gestartet werden drfen.
-       Liste soll dabei nicht dauernd repainted werden. */
-    sa.mLevelListe->setAutoUpdate(false);
-
-    int glnr = mSpielerZahl > 1; // Welche Gewonnen-Liste?
-    int leanz = ld->getLevelAnz();
-
-    /* Level durchgehen */
-    bool alle_gewonnen = true;
-    int lenr;
-    for (lenr = 1; lenr <= leanz; lenr++) {
-      /* In Liste einfgen */
-      __String lena = ld->getLevelName(lenr);
-      __String zei;
-      zei.sprintf("%d. %s", lenr, lena.data());
-      sa.mLevelListe->insertItem((QString) zei);
-
-      /* Schon mal gewonnen? */
-      __String intlena= ld->getIntLevelName(lenr);
-      if (!mGewonneneLevel[glnr].contains(intlena)) {
-	/* Der Level war noch nicht gewonnen; also keine weiteren 
-	   in die Liste */
-	alle_gewonnen = false;
-	break;
-      }
-    }
-    sa.mLevelListe->setAutoUpdate(true);
-    sa.mLevelListe->repaint();
-
-    /* Versuchen zu erraten, welchen Level der User haben will */
-    sa.mLevelListe->setCurrentItem((
-				    mLevelNr == 0 ?
-				    alle_gewonnen ? 1 : lenr
-				    : mLevelNr) - 1);
-			
-						
-    if (sa.exec()) {
-      mLevelNr = sa.mLevelListe->currentItem() + 1;
-      ASSERT(mLevelNr >= 1);
+    if (gPrefs->startAtDialog(this, mSpielerZahl > 1, mLevelNr))
       startSpiel();
-    }
     break;
   }
 
   case MENU_pause_game: {
-    ASSERT(mGModus == gmodus_spiel);
+    CASSERT(mGModus == gmodus_spiel);
 
     setGModus(gmodus_pause);
     for (int i = 0; i < mSpielerZahl; i++)
@@ -296,7 +334,7 @@ void Cuyo::MenuAufruf(int id) {
     if (mSpielerZahl != 1) {
       setSpielerZahl(1);
       mLevelNr = 0;
-				/* Ist restart level noch enabled? */
+      /* Ist restart level noch enabled? */
       aktualisiereMenuEnabled();
     }
     break;
@@ -305,13 +343,13 @@ void Cuyo::MenuAufruf(int id) {
     if (mSpielerZahl != 2 || mGegenKI) {
       setSpielerZahl(2);
       mLevelNr = 0;
-				/* Ist restart level noch enabled? */
+      /* Ist restart level noch enabled? */
       aktualisiereMenuEnabled();
     }
     break;
 
   case MENU_KI_player:
-    if (mSpielerZahl != spz_ki) {
+    if (!mGegenKI) {
       setSpielerZahl(spz_ki);
       mLevelNr = 0;
       /* Ist restart level noch enabled? */
@@ -320,24 +358,25 @@ void Cuyo::MenuAufruf(int id) {
     break;
 					
   case MENU_preferences: {
-    Prefs p(this);  /* this = Fenster, ber dem der Dialog liegt */
-			
-    /* Aktuelle Einstellungen in den Dialog schreiben */
-    for (int i = 0; i < anz_spieler_in_dlg; i++)
-      for (int j = 0; j < 4; j++)
-	p.mTasten[i][j]->setTaste(mTasten[i][j]);
-    //p.mTaste1Spieler->setChecked(mTaste1Spieler);
-			
-    /* Dialog ausfhren */
-    if (p.exec()) {
-      /* OK geklickt => Einstellungen wieder aus Dialog holen */
-      for (int i = 0; i < anz_spieler_in_dlg; i++)
-	for (int j = 0; j < 4; j++)
-	  mTasten[i][j] = p.mTasten[i][j]->getTaste();
-      //mTaste1Spieler = p.mTaste1Spieler->isChecked();
-						
-      schreibPreferences();
-    }
+    /* Wenn der Benutzer umschaltet, welche Level aktiv sind, soll
+       der Level mit dem selben Namen markiert bleiben, und nicht
+       mit der selben Nummer... */
+    __String levelname;
+    if (mLevelNr)
+      levelname = ld->getIntLevelName(mLevelNr);
+
+    gPrefs->preferencesDialog(this);
+
+    /* So, und jetzt wieder aus dem Namen die Nummer kriegen.
+       (Ist evtl. 0, wenn der Level nicht mehr existiert) */
+    if (mLevelNr)
+      mLevelNr = ld->getLevelNr(levelname);
+
+    /* Wenn jetzt pltzlich mLevelNr == 0 ist, darf "Restart Level"
+       nicht mehr gehen. */
+    aktualisiereMenuEnabled();
+
+
     break;
   }
   
@@ -356,20 +395,25 @@ void Cuyo::startSpiel() {
   if (mGModus != gmodus_kein_spiel)
     stopSpiel();
 
+  /* Damit der Level nicht schon beim Start angehalten ist oder schnell
+     luft */
+  mEinzelschritt = false;
+  mSchnellvorlauf = false;
+
   if (mSpielerZahl == 1)
     mSpielfeld[1]->setText(_("Your ad here!\n\n"));
-			
+
   for (int i = 0; i < max_spielerzahl; i++)
     setPunkte(i, 0);
 
   try {
     if (!startLevel())
-      throw Fehler(_("There are no (more?) levels in level.descr."));
+      throw Fehler(_("There are no (more?) (working?) levels."));
 
   } catch (Fehler f) {
-    fprintf(stderr, f.mText.data());
+    fprintf(stderr, "%s\n", f.getText().data());
     for (int i = 0; i < mSpielerZahl; i++)	
-      mSpielfeld[i]->setText(f.mText);
+      mSpielfeld[i]->setText(f.getText(), true, true);
 
     /* Sicherheitshalber...: */
     stopSpiel();
@@ -380,23 +424,39 @@ void Cuyo::startSpiel() {
   /* Timer erzeugen und verbinden... */	
   mTimer = new QTimer(this);
 	
-  connect( mTimer, SIGNAL(timeout()), SLOT(spielSchritt()) );
+  connect(mTimer, SIGNAL(timeout()), SLOT(spielSchritt()));
 
-  mTimer->start( 80 );
+  mTimer->start(80);
 
   aktualisiereMenuEnabled();
 }
 
-/** stoppt das Spiel sofort (egal, ob grad ein Level luft oder nicht) */
-void Cuyo::stopSpiel(){
-  if (mGModus == gmodus_kein_spiel)
-    return;
 
-  stopLevel(true);
-			
-  delete mTimer;
+/** stoppt das Spiel sofort (egal, ob grad ein Level luft oder nicht) */
+void Cuyo::stopSpiel(bool wegen_fehler/* = false*/) {
+  /* Evtl. wird stopSpiel() aufgerufen, wenn das Spiel noch gar nicht
+     richtig fertig gestartet wurde; dann wurde mGModus vielleicht noch
+     nicht gesetzt, aber es soll trotzdem was gestoppt werden... */
+  /*if (mGModus == gmodus_kein_spiel)
+    return;*/
+
+  stopLevel(true, wegen_fehler);
+
+  /* Evtl. wird das Spiel zu einem Zeitpunkt gestoppt, zu dem
+     noch gar kein Timer existiert... */
+  if (mTimer)	
+    delete mTimer;
+  mTimer = 0;
 	
   setGModus(gmodus_kein_spiel);
+
+  ld->ladKeinenLevel();
+  
+  /* berall ndert sich die Textfarbe... */
+  for (int i = 0; i < max_spielerzahl; i++) {
+    mPunkteFeld[i]->update();
+    mSpielfeld[i]->update();
+  }
 }
 
 
@@ -404,9 +464,12 @@ void Cuyo::stopSpiel(){
 
 
 
-/** setzt Level hoch u. . */
+/** Die Haupt-Spielschritt-Routine, whrend das Spiel luft. Wird direkt
+    von QT aufgerufen. Ruft alle anderen spielschritt()-Routinen auf. */
 void Cuyo::spielSchritt() {
-  ASSERT(mGModus != gmodus_kein_spiel);
+  CASSERT(mGModus != gmodus_kein_spiel);
+  
+  mInTime = true;
 
   /* Zeitlupen-Debug-Modus? */
   if (mZeitlupe) {
@@ -431,12 +494,21 @@ void Cuyo::spielSchritt() {
     } else
       return;
   }
+  
+  /* Einzelschritt-Modus. Nach diesem Schritt gleich wieder auf
+     Tastendruck warten. */
+  if (mEinzelschritt) {
+    mWarteAufTaste = true;
+  }
 
   /* Ist grade eine Pause beendet? */
   if (mGModus == gmodus_pause) {
     setGModus(gmodus_spiel);
-    for (int i = 0; i < mSpielerZahl; i++)
+    for (int i = 0; i < mSpielerZahl; i++) {
+      /* Das ist nicht ganz richtig: Wenn da grad ein Text am Blinken
+         war, ist er jetzt erst mal weg. */
       mSpielfeld[i]->setText("");
+    }
   }
 
 
@@ -449,6 +521,7 @@ void Cuyo::spielSchritt() {
     stopLevel(false);
  		
     /* neuer Level */
+    
     mLevelNr++;
     try {
       if (!startLevel()) {
@@ -456,16 +529,17 @@ void Cuyo::spielSchritt() {
         stopSpiel();
         for (int i = 0; i < mSpielerZahl; i++) {
   	__String s;
-  	s.sprintf(_("***\nYou won!!!\n\nScore: %d\n***\n\n"), mPunkte[i]);
+  	s.sprintf(mSpielerZahl == 2 && mPunkte[i] > mPunkte[1-i] ? 
+		    _("***\nYou won even a bit more!!!\n\nScore: %d\n***\n\n") :
+		    _("***\nYou won!!!\n\nScore: %d\n***\n\n")  ,
+	          mPunkte[i]);
   	mSpielfeld[i]->setText(s);
         }
         mLevelNr = 0; // Jetzt kann man nicht mehr Restart last level
+        aktualisiereMenuEnabled();
       }
-    } catch (Fehler f) {
-      fprintf(stderr, f.mText.data());
-      for (int i = 0; i < mSpielerZahl; i++)	
-        mSpielfeld[i]->setText(f.mText);
-  
+    } catch (Fehler fe) {
+      printFehler(mSpielerZahl, _("Could not start level:\n"), fe);
       stopSpiel();
     }
     return;
@@ -478,52 +552,85 @@ void Cuyo::spielSchritt() {
     setGModus(gmodus_spiel);
   }
 
+
+  /* Anfang der Schnellvorlauf-Schleife */
+  int i_sv = 0;
+schnellvorlauf:  
+
+
   /* Ok, hier kommt der eigentliche Spielschritt */
   if (mGModus != gmodus_bonus_animation) {
 
+    /* Evtl. Signale abfangen, um die Logdatei abspeichern zu knnen. */
+    signaleAn();
+    
     try {
+    
+      /* Den Schritt aufnehmen bzw. abspielen. Beim Abspielen werden
+         hier auch ggf. Tastendrcke abgespielt. */
+      Aufnahme::recSchritt(mSpielfeld);
       
-      /* Luft eine hifea-Animation oder ist normales Spiel? */
-      if (mSpielfeld[0]->getHifea() ||
-          mSpielerZahl == 2 && mSpielfeld[1]->getHifea()) {
-        /* hifea-Animation */
-        for (int i = 0; i < mSpielerZahl; i++)
-          mSpielfeld[i]->hifeaSchritt();
-          
-      } else {
-        /* Normales Spiel */
-        ld->spielSchritt(); // Fr Synchron-Animationen...
-        
-        for (int i = 0; i < mSpielerZahl; i++)
-          mSpielfeld[i]->spielSchritt();
+      /* Erst mal den normalen Spielschritt... in einer Gleichzeit, fr
+         evtl. auftretende Events. (Init-Events nicht nicht dabei; die
+	 werden erst beim darauffolgenden animiere() behandelt.) */
+      Blop::beginGleichzeitig();
+      for (int i = 0; i < mSpielerZahl; i++)
+        mSpielfeld[i]->spielSchritt();
+      Blop::endGleichzeitig();
+
+      /* Animationen und Grafik-Ausgabe */
+      animiere();
+
+      /* Ggf. auch einen Spielschritt fr die KI */
+      if (mGegenKI)
+        mKI->spielSchritt();
+
+    } catch (Fehler fe) {
+      signaleAus();
+      bool log_gespeichert = false;
+
+      try {
+        Aufnahme::speichern(getLogName());
+	log_gespeichert = true;
+      } catch (Fehler) {}
+
+      /* Im Debug-Modus soll das Spielfeld weiter angezeigt werden,
+         wenn ein Fehler whrend des Spiels passiert. */
+      stopSpiel(gDebug);
+      
+      /* Wenn das Speichern der log-Datei geklappt hat, soll ggf.
+         die "send log to..."-Meldung ausgegeben werden. */
+      printFehler(mSpielerZahl, _("Error during the game:\n"), fe,
+                  log_gespeichert);
 
-        /* Ggf. auch einen Spielschritt fr die KI */
-        if (mGegenKI)
-          mKI->spielSchritt();
-      }
-		
-        
-    } catch (Fehler f) {
-      /* Im Spielschritt ein Fehler aufgetreten? */
-      stopSpiel();
-      fprintf(stderr, _("An error occured:\n%s\n"), f.mText.data());
-				
-      for (int i = 0; i < mSpielerZahl; i++) {
-	__String s;
-	s.sprintf(_("An error occured:\n%s\n"), f.mText.data());
-	mSpielfeld[i]->setText(s);
-      }
       return;
     }
+    	      
+    signaleAus();
   }
 
-  /* Wurde grade das restliche Gras vernichtet? Dann warten wir jetzt nur noch
-     auf einen guten Moment zum beenden. mGrasDaAnz < 0 kann nicht wirklich
-     vorkommen. Aber im Debug-Modus vielleicht, wenn das Gras knstlich
-     auf 0 gesetzt wurde. */
-  if (mGModus == gmodus_spiel && mGrasDaAnz <= 0)
+  if (mGModus == gmodus_spiel) {
+    /* Wurde grade das restliche Gras vernichtet? */
+    for (int i = 0; i < mSpielerZahl; i++)
+      if (mSpielfeld[i]->istGrasDa())
+        goto noch_gras_da;
+      
+    /* Dann warten wir jetzt nur noch auf einen guten Moment zum beenden. */
     setGModus(gmodus_warte_level);
-			
+      
+noch_gras_da:;
+  }
+
+
+
+  /* Ende der Schnellvorlauf-Schleife. Evtl. gleich zwei weitere
+     Spielschritte ausfhren. */
+  if (mSchnellvorlauf && i_sv < 3) {
+    i_sv++;
+    goto schnellvorlauf;
+  }
+
+  
 			
   if (mGModus == gmodus_warte_level ||
       mGModus == gmodus_warte_tot) {
@@ -544,15 +651,10 @@ void Cuyo::spielSchritt() {
 	/* Hier darf noch nicht stopLevel() aufgerufen werden, weil sonst
 	   das Spielfeld nicht mehr angezeigt wird */
 	
-	/* Gewonnen. Level als gewonnen abspeichern. */
-	/* Ein- oder Mehrspielermodus? */
-	int glnr = mSpielerZahl > 1;
-	__String intlena = ld->getIntLevelName(mLevelNr);
-	if (!mGewonneneLevel[glnr].contains(intlena.data())) {
-	  /* Level war bisher noch nie gewonnen; also in Liste einfgen */
-	  mGewonneneLevel[glnr].append(intlena);
-	  schreibPreferences();
-	}
+	/* Gewonnen. Level als gewonnen abspeichern. Aber nicht im
+	   Debug-Modus. */
+	if (!gDebug)
+	  gPrefs->schreibGewonnenenLevel(mSpielerZahl > 1, mLevelNr);
 				
 	/* Jetzt kommt die Zeitbonus-Animation */
 	setGModus(gmodus_bonus_animation);
@@ -576,36 +678,72 @@ void Cuyo::spielSchritt() {
 
   if (mGModus == gmodus_bonus_animation) {
     /* Hetzrand runterrutschen lassen */
-    bool ba_fertig;
+    bool ba_fertig = true; // "= true" nur um eine Warnung zu ersparen
 		
     mZeitBonus += punkte_fuer_zeitbonus;
 
     for (int i = 0; i < mSpielerZahl; i++) {
+
+      /* Neuen Text (mit neuer Punkt-Zahl) schreiben */
+      __String s;
+      s.sprintf(_("Level %s complete!\n\nTime Bonus: %d\nScore: %d\n\n"),
+		ld->mLevelName.data(), mZeitBonus, mPunkte[i]);
+      /* Kein update-Event; bei bonusSchritt wird eh neu gemalt. */
+      mSpielfeld[i]->setText(s, false);
+
       /* Rand runterrutschen lassen */		
       ba_fertig = mSpielfeld[i]->bonusSchritt();
  		
       /* Punkte bekommen */
       setPunkte(i, mPunkte[i] + punkte_fuer_zeitbonus);
-			
-      /* Neuen Text (mit neuer Punkt-Zahl) schreiben */
-      __String s;
-      s.sprintf(_("Level %s complete!\n\nTime Bonus: %d\nScore: %d"),
-		ld->mLevelName.data(), mZeitBonus, mPunkte[i]);
-      mSpielfeld[i]->setText(s);
     }
 				
     /* Unten angekommen? */
     if (ba_fertig) {
       setGModus(gmodus_bonus_warte);
+
+      for (int i = 0; i < mSpielerZahl; i++) {
+        __String s;
+        s.sprintf(_("Level %s complete!\n\nTime Bonus: %d\nScore: %d\n\nSpace = Continue"),
+	  	ld->mLevelName.data(), mZeitBonus, mPunkte[i]);
+        mSpielfeld[i]->setText(s);
+      }
  			
       mWarteAufTaste = true;
-      mWarteTimeout = nachbonus_wartezeit;
+      /* Kein Warte-Timeout mehr. Einfach Leertaste drcken */
+      //mWarteTimeout = nachbonus_wartezeit;
     } // Ende: if Bonus-Animation fertig
   } // Ende: if Bonus-Animation
 }
 
 
 
+
+/** Wird von startLevel() und von spielSchritt() aufgerufen. Lsst
+    smtliche Blops animieren und updatet die Grafik. */
+void Cuyo::animiere() {
+
+  /* Vielleicht htten gerne einige Blops noch ein Init-Event... */
+  Blop::sendeFehlendeInitEvents();
+
+  /* Erst mal das globale Blop ausfhren. */
+  ld->spielSchritt();
+  
+  /* Jetzt alle anderen... in *einer* Gleichzeit */
+  Blop::beginGleichzeitig();
+  
+  /* Alle Grafiken lschen */
+  Blop::lazyInitStapel();
+  
+  /* Alles neu malen  */
+  for (int i = 0; i < mSpielerZahl; i++)
+    mSpielfeld[i]->animiere(); // updatet auch die Grafik
+    
+  Blop::endGleichzeitig();
+}
+
+
+
 /**  */
 void Cuyo::neuePunkte(bool reSp, int pt){
   if (reSp) setPunkte(1, mPunkte[1] + pt);
@@ -613,19 +751,10 @@ void Cuyo::neuePunkte(bool reSp, int pt){
 }
 
 
-/**  */
-void Cuyo::grasGeplatzt() {
-  mGrasDaAnz--;
-  ASSERT(mGrasDaAnz >= 0);
-}
-
-
 /**  */
 void Cuyo::setPunkte(int sp, int pu){
-  __String s;
   mPunkte[sp] = pu;
-  s.sprintf("%d", pu);
-  punkteLabel[sp]->setText(s);
+  mPunkteFeld[sp]->setPunkte(pu);
 }
 
 
@@ -637,35 +766,62 @@ void Cuyo::setPunkte(int sp, int pu){
     den Level gar nicht mehr gibt. Throwt bei Fehler. */
 bool Cuyo::startLevel() {
 
-  int i;
-
   /* So viele Level gibt's doch gar nicht */
   if (mLevelNr > ld->getLevelAnz())
     return false;
 
   /* Level-Daten laden */
-  for (i = 0; i < mSpielerZahl; i++) {
+  for (int i = 0; i < mSpielerZahl; i++) {
     __String s;
     s.sprintf(_("Score: %d\n\nLoading Level %d...\n\n"),
 	      mPunkte[i], mLevelNr);
-    mSpielfeld[i]->setText(s, true);
+    mSpielfeld[i]->setText(s);
+    /* Das false besagt: Nicht erst Lschen */
+    mSpielfeld[i]->repaint(0, 0,
+              mSpielfeld[i]->width(), mSpielfeld[i]->height(), false);
   }
   ld->ladLevel(mLevelNr, mSpielerZahl);
+  
+  for (int i = 0; i < max_spielerzahl; i++) {
+    /* Punkte-Anzeige jetzt in passender Farbe */
+    mPunkteFeld[i]->update();
+  }
+  /* Und das "Your ad here" auch... */
+  if (mSpielerZahl == 1)
+    mSpielfeld[1]->update();
 	
-  /* Gesamtgrasmenge berechnen */
-  mGrasDaAnz = ld->getGrasAnzahl() * mSpielerZahl;
-	
+  /* Aufnehmen / Abspielen starten */
+  Aufnahme::init(mAbspielen, mGegenKI ? spz_ki : mSpielerZahl);
+
+  /***** Allen Leuten erzhlen, dass jetzt ein Level anfngt. *****/
+  
+  /* Fr den Global-Blop... */
+  ld->startLevel();
+
   /* Hier ist die Stelle, wo wir darauf gucken, wie viele Spieler mitspielen;
      nur fr so viele Spieler wird das Spiel wirklich gestartet */
-  for (i = 0; i < mSpielerZahl; i++) {
+  for (int i = 0; i < mSpielerZahl; i++) {
+    
     mSpielfeld[i]->startLevel();
     __String s;
-    s.sprintf(_("Score: %d\n\nLevel %d\n%s\n%d blops explode\n%s\n%s\n\nSpace = Start"),
-	    mPunkte[i], mLevelNr, ld->mLevelName.data(), ld->mPlatzAnzahl,
+    s.sprintf(_("Score: %d\n\n"
+                "Level %d\n%s\nby %s\n\n"
+                "%d blops explode\n%s\n"
+		"%s\n\n"
+		"Space = Start"),
+	    mPunkte[i],
+	    mLevelNr, ld->mLevelName.data(), ld->mLevelAutor.data(),
+	    ld->mPlatzAnzahl,
 	    ld->mGrasBeiKettenreaktion ? "Chain reaction necessary\n" : "",
 	    ld->mBeschreibung.data());
+    /* Kein update-Event verschicken; beim animiere()-Aufruf wird
+       eh neu gemalt. */
     mSpielfeld[i]->setText(s, false);
   }
+  
+  /* Alle Blops einmal animieren, damit sie wissen, wie sie aussehen.
+     Dabei werden auch die ganzen Init-Events verschickt. */
+  animiere();
 
   /* Auch die KI mchte wissen, wenn ein neuer Level anfngt. */
   if (mGegenKI)
@@ -684,10 +840,10 @@ bool Cuyo::startLevel() {
     getan werden muss (ohne Animation, d. h. entweder
     ist die Animation schon vorbei oder es gibt halt keine).
     Wenn malen = true ist, Bildschirm sofort updaten */
-void Cuyo::stopLevel(bool malen) {
+void Cuyo::stopLevel(bool malen, bool wegen_fehler /*= false*/) {
 			
   for (int i = 0; i < mSpielerZahl; i++)
-    mSpielfeld[i]->stopLevel(malen);
+    mSpielfeld[i]->stopLevel(malen, wegen_fehler);
 }
 
 
@@ -697,7 +853,7 @@ void Cuyo::stopLevel(bool malen) {
 /** liefert true, wenn das Spiel normal luft, false
 		wenn das Spiel am zuende gehen ist */
 bool Cuyo::spielLaeuft() {
-  ASSERT(mGModus != gmodus_kein_spiel);
+  CASSERT(mGModus != gmodus_kein_spiel);
   return mGModus == gmodus_spiel || mGModus == gmodus_pause;
 }
 
@@ -717,6 +873,18 @@ void Cuyo::tasteWeiter() {
   mWarteTimeout = 0;
 }
 
+
+/** Liefert ein Spielfeld zurck. */
+Spielfeld * Cuyo::getSpielfeld(bool reSp) {
+  return mSpielfeld[reSp];
+}
+
+/** Liefert zurck, ob die Zeit fr Grafik grade reicht */
+bool Cuyo::getInTime() const {
+  return mInTime;
+}
+
+
 /** Liefert die Anzahl der Mitspieler zurck. */
 bool Cuyo::getSpielerZahl() const {
   return mSpielerZahl;
@@ -728,32 +896,18 @@ void Cuyo::keyPressEvent(QKeyEvent * e) {
   int k = e->key();
   bool acc = false;
   if (mGModus == gmodus_spiel) {
-    for (int i = 0; i < max_spielerzahl; i++) {
-  	
+    int sp, t;
+    if (gPrefs->getTaste(k, sp, t)) {
       /* Im 1-Spieler-Modus und im KI-Modus alle Tastendrcke an
-	 Spieler 1 senden; sonst Tasten an die verschiedenen Spieler
-	 verteilen */
-      int j;
-      if (mSpielerZahl == 1 || mGegenKI) j = 0; else j = i;
-  		
-      if (k == mTasten[i][taste_links]) {
-	mSpielfeld[j]->tasteLinks();
-	acc = true;
-      }
-      if (k == mTasten[i][taste_rechts]) {
-	mSpielfeld[j]->tasteRechts();
-	acc = true;
-      }
-      if (k == mTasten[i][taste_dreh]) {
-	mSpielfeld[j]->tasteDreh();
-	acc = true;
-      }
-      if (k == mTasten[i][taste_fall]) {
-	mSpielfeld[j]->tasteFall();
-	acc = true;
-      }
+	 Spieler 1 senden: */
+      if (mSpielerZahl == 1 || mGegenKI) sp = 0;
+
+      Aufnahme::recTaste(sp, t);
+      mSpielfeld[sp]->taste(t);
+      acc = true;
     }
   } // if spiel luft
+  
   if (mWarteAufTaste)
     if (k == Key_Space || k == Key_Return || k == Key_Enter) {
       tasteWeiter();
@@ -774,33 +928,93 @@ void Cuyo::keyPressEvent(QKeyEvent * e) {
 	return; 
       }
 
-      if (a == 'f') {
-	mFlag = !mFlag;
-	fprintf(stderr, "Debug: Flag = %d\n", mFlag);
+      if (a == 'e') {
+	mEinzelschritt = !mEinzelschritt;
+	fprintf(stderr, "Debug: Einzelschritt = %d\n", mEinzelschritt);
+        if (!mEinzelschritt) {
+          /* Nicht noch ein letztes Mal auf Taste warten... */
+          mWarteAufTaste = false;
+        }
+	return;
+      }
+
+      if (a == 'L') {
+	fprintf(stderr, "Debug: Logdatei laden\n");
+        Aufnahme::laden(getLogName());
+	setSpielerZahl(Aufnahme::getSpZ());
+	__String lna = Aufnahme::getLevelName();
+        int leanz = ld->getLevelAnz();
+
+	mLevelNr = 0;
+        /* Level durchgehen */
+        for (int lenr = 1; lenr <= leanz; lenr++) {
+          if (ld->getIntLevelName(lenr) == lna) {
+	    mLevelNr = lenr;
+	  }
+	}
+	
+	fprintf(stderr, "Level %s (%d)\n",
+	        lna.data(), mLevelNr);
+	fprintf(stderr, "Zum Abspielen ctrl-r drcken.\n");
+	mAbspielen = true;
+	
+        aktualisiereMenuEnabled();
+      }
+
+      if (a == 'l') {
+	mAbspielen = !mAbspielen;
+	fprintf(stderr, "Debug: Log abspielen = %d\n", mAbspielen);
 	return;
       }
 
       if (a == 'g') {
-	mGrasDaAnz = 0;
-	fprintf(stderr, "Debug: Gewinnen\n");
+        /* Hoffentlich darf man gModus einfach so ndern... */
+        
+        if (mGModus == gmodus_spiel) {
+          setGModus(gmodus_warte_level);
+	  fprintf(stderr, "Debug: Gewinnen\n");
+        }
 	return;
       }
 
       if (a == 'h') {
 	fprintf(stderr, "Debug: Hilfe\n");
 	fprintf(stderr, "  b: Bekomme Graue\n");
-	fprintf(stderr, "  f: Flag an/aus\n");
+	fprintf(stderr, "  e: Einzelschrittmodus an/aus\n");
 	fprintf(stderr, "  g: Gewinnen\n");
 	fprintf(stderr, "  h: Hilfe\n");
+	fprintf(stderr, "  L: Log laden\n");
+	fprintf(stderr, "  l: Log abspielen an/aus\n");
+	//fprintf(stderr, "  n: Neue Animationen testen\n");
 	fprintf(stderr, "  r: Reload levelconf\n");
+	fprintf(stderr, "  S: Log speichern\n");
 	fprintf(stderr, "  t: Test Rber-Reihe an/aus\n");
+	fprintf(stderr, "  v: Schnellvorlauf an/aus\n");
 	fprintf(stderr, "  z: Zeitlupe an/aus\n");
 	return;
       }
+      /*
+      if (a == 'n') {
+
+      }*/
 
       if (a == 'r') {
-	ld->ladLevelConf();
-	fprintf(stderr, "Debug: Reload levelconf\n");
+        if (mGModus == gmodus_kein_spiel) {
+	  fprintf(stderr, "Debug: Reload levelconf\n");
+          try {
+   	    ld->ladLevelConf();
+	  } catch (Fehler fe) {
+	    printFehler(2, "Could not reload the level description file:\n",
+	        	fe);
+	  }
+	} else
+	  fprintf(stderr, "Debug: *Nicht* reload levelconf whrend Spiel luft.\n");
+	return;
+      }
+
+      if (a == 'S') {
+	fprintf(stderr, "Debug: Logdatei speichern\n");
+        Aufnahme::speichern(getLogName());
 	return;
       }
 
@@ -810,6 +1024,12 @@ void Cuyo::keyPressEvent(QKeyEvent * e) {
 	return;
       }
 
+      if (a == 'v') {
+	mSchnellvorlauf = !mSchnellvorlauf;
+	fprintf(stderr, "Debug: Schnellvorlauf = %d\n", mSchnellvorlauf);
+	return;
+      }
+
       if (a == 'z') {
 	mZeitlupe = !mZeitlupe;
 	fprintf(stderr, "Debug: Zeitlupe = %d\n", mZeitlupe);
@@ -822,67 +1042,9 @@ void Cuyo::keyPressEvent(QKeyEvent * e) {
   }
 }
 
-/** Ldt die Preferences aus wo-auch-immer-sie-abgespeichert-werden
-in die Variablen. */
-void Cuyo::liesPreferences(){
-	
-  int i;
-  ConfigDatei prd(getPrefsName());
-	
-  /* Tastenbelegung */
-	
-  /* Default-Tasten */
-  int dt[max_spielerzahl][4] = {{Key_A, Key_D, Key_W, Key_S},
-    {Key_Left, Key_Right, Key_Up, Key_Down}};
-  char tn[4][9] = {"left", "right", "turn", "down"};
-																	
-  for (i = 0; i < max_spielerzahl; i++) {
-    __String s;
-    s.sprintf("keys %d", i + 1);
-    prd.setAbschnitt(s);
-    for (int j = 0; j < 4; j++)
-      mTasten[i][j] = prd.getZahlEintrag(tn[j], dt[i][j]);
-  }
-
-  /* Gewonnene Level */
-  prd.setAbschnitt(0);
-  for (i = 0; i < 2; i++) {
-    prd.getListenEintrag(i ? "wonMultiPlayer" : "wonOnePlayer", mGewonneneLevel[i]);
-    /* Debug-Ausgabe: */
-    /*__String t = prd->readEntry(i ? "wonMultiPlayer" : "wonOnePlayer");
-      printf("blub = %s\n", t.data());*/
-  }
-}
-
-/** Schreibt die Preferences nach wo-auch-immer-sie-abgespeichert-werden. */
-void Cuyo::schreibPreferences(){
-
-  QFile f(getPrefsName());
-  f.open(IO_WriteOnly);
-  QTextStream ts(&f);
-  ts << __String("# cuyo Preferences File\n");
-  ts << __String("\n");
-  for (int i = 0; i < 2; i++) {
-    ts << __String(i ? "wonMultiPlayer" : "wonOnePlayer") << __String("=");
-    __String lena;
-    for (lena = mGewonneneLevel[i].first(); lena != 0;
-	 lena = mGewonneneLevel[i].next())
-      ts << lena << __String(",");
-    ts << __String("\n");
-  }
-  ts << __String("\n");
-	
-  char tn[4][9] = {"left", "right", "turn", "down"};
-  for (int i = 0; i < max_spielerzahl; i++) {
-    ts << __String("[keys ") << (i + 1) << __String("]\n");
-    for (int j = 0; j < 4; j++)
-      ts << __String(tn[j]) << __String("=") << mTasten[i][j] << __String("\n");
-    ts << __String("\n");
-  }
-	
-}
-
-/** Setzt die Anzahl der Spieler. Auch in der Menleiste. */
+/** Setzt die Anzahl der Spieler. Auch in der Menleiste. 
+    Bei a == spz_ki wird auf ki-Modus geschaltet. (spz_ki wird in
+    aufnahme.h definiert, weil es dort auch gebraucht wird.) */
 void Cuyo::setSpielerZahl(int a){
   mGegenKI = a == spz_ki;
   mSpielerZahl = mGegenKI ? 2 : a;
@@ -913,17 +1075,41 @@ void Cuyo::aktualisiereMenuEnabled() {
   mSpielMenu->setItemEnabled(MENU_preferences, ks);
 }
 
-/** Liefert den Namen und Pfad der Prefs-Datei zurck
-    ($HOME/.cuyo) */
-__String Cuyo::getPrefsName() const{
+
+
+/** Gibt die Fehlermeldung bestehend aus t und fe aus:
+    Sowohl als Text im Cuyo-Fenster
+    (bei anz_sp vielen Spielern) als auch auf stderr. 
+    mitLog wird an fe.getText() weitergegeben. (D. h.: soll
+    ggf. die Send-Log-Meldung ausgegeben werden?) */
+void Cuyo::printFehler(int anz_sp, __String t, Fehler fe,
+                       bool mitLog /*= false*/) {
+  t += fe.getText(mitLog);
+  fprintf(stderr, "%s\n", t.data());
+  for (int i = 0; i < anz_sp; i++) {
+    /* zweites true = Kleine Schrift... */
+    mSpielfeld[i]->setText(t, true, true);
+  }
+}
+
+
+
+
+/***** Die nachfolgenden Funktionen gehren nicht zum Cuyo-Objekt *****/
+
+
+
+/** Liefert den Namen und Pfad der Log-Datei zurck
+    ($HOME/cuyo.log) */
+__String getLogName() {
   char * ho = getenv("HOME");
   if (!ho) {
     /* Unter Windows zum Beispiel... */
-    fprintf(stderr, _("Warning: Env-Variable $HOME not found. Using the current cirectory for .cuyo"));
-    return ".cuyo";
+    fprintf(stderr, _("Warning: Env-Variable $HOME not found. Using the current cirectory for cuyo.log"));
+    return "cuyo.log";
   }
   if (ho[strlen(ho) - 1] == '/')
-    return __String(ho) + ".cuyo";
+    return __String(ho) + "cuyo.log";
   else
-    return __String(ho) + "/.cuyo";
+    return __String(ho) + "/cuyo.log";
 }
