@@ -24,6 +24,12 @@
 
 #include "inkompatibel.h"
 
+/* Versucht die Datei na zu laden und in QImage zu speichern.
+   Versucht auerdem, die Datei na.gz zu lagen.
+   Liefert false, wenn keine der Dateien existiert.
+   Throwt, wenn's beim Laden einen Fehler gibt.
+   (Falls die QT-Lad-Routine verwendet wird, kann nicht versucht werden,
+   die .gz-Datei zu laden.) */
 bool ladXPM(QImage & im, __String na);
 
 
